/* hanwin.c: Tcl (Tk) liaison functions for Hanzi Master.  These are
   called by Tk event handlers to generate various lists of characters. */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.  Please see the file "COPYING"
   for details.  If you have not received it along with this program,
   please write to the Free Software Foundation, Inc., 59 Temple Place,
   Suite 330, Boston, MA 02111-1307, USA. */

#include "hanzim.h"

char	simpFlag;

/* sameRadList() returns a list of character,pronunciation,meaning triples
   (or actually character index, font representation, pinyin font representat-,
    ion, alphanumeric, and meaning quintuples) comprising all characters
   containing same radical as a character given as argument; currently nothing
   done about multiple pronunciations, although they are in the database */
int
sameRadList(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  int		typ, i,
		givench,
		pchidx,
		flevel,
		gradidx;
  radicalst	*grad;
  characterst	*pch;
  char		resultstr[120];

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    givench = atoi(argv[1]);
    flevel = atoi(argv[2]);
    simpFlag = strcmp(argv[3],"fanti");
    }

  /* return all characters having this radical in any position */
  Tcl_ResetResult(interp);
  gradidx = chartbl[givench].radidx;
  if (gradidx >= 0) {
    grad = &(radtbl[chartbl[givench].radidx]);
    for (typ=0;typ<4;typ++) {
      for (i=0;i<grad->chars[typ].n_entries;i++) {
	pch = &(chartbl[grad->chars[typ].ent[i]]);
	if ((pch->radtype == typ) && (pch->freq < flevel) &&
	    ((pchidx = pch - chartbl) != givench)) {
	  construct_char_entry(resultstr,pchidx);
	  Tcl_AppendElement(interp,resultstr);
	  }
	}
      }
    }
  return TCL_OK;
}


/* sameRemList() returns a list similar to above except for same remainder
   instead of same radical */
int
sameRemList(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
  int		i,
		givench,
    		givenrem,
		flevel,
    		cidx;
  idxlist	*ro;
  char		resultstr[120];

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    givench = atoi(argv[1]);
    flevel = atoi(argv[2]);
    simpFlag = strcmp(argv[3],"fanti");
    }

  givenrem = chartbl[givench].remainder;

  /* clear last and return new result */
  Tcl_ResetResult(interp);
  if (givenrem >= 0) {
    /* first add remainder itself, then list of others */
    construct_char_entry(resultstr,givenrem);
    Tcl_AppendElement(interp,resultstr);
    ro = &(chartbl[givenrem].remOf);
    for (i=0;i<ro->n_entries;i++) {
      cidx = ro->ent[i];
      if ((chartbl[cidx].freq < flevel) && (cidx != givench)) {
	construct_char_entry(resultstr,cidx);
	Tcl_AppendElement(interp,resultstr);
	}
      }
    }

  return TCL_OK;
}


/* samePronList() same as previous two except for matching pronunciation */
int
samePronList(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
  int		i,
		givench,
    		pchidx,
		flevel;
  pinyinst	*gpin;
  characterst	*pch;
  char		resultstr[120];

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    givench = atoi(argv[1]);
    flevel = atoi(argv[2]);
    simpFlag = strcmp(argv[3],"fanti");
    }

  /* return all characters on the list of pinyin which is most frequent
     pronunciation of this character */
  Tcl_ResetResult(interp);
  gpin = &(pinyintbl[chartbl[givench].pinyins.ent[0]]);
  for (i=0;i<gpin->chars.n_entries;i++) {
    pch = &(chartbl[gpin->chars.ent[i]]);
    if ((pch->freq < flevel) && ((pchidx = pch - chartbl) != givench)) {
      construct_char_entry(resultstr,pchidx);
      Tcl_AppendElement(interp,resultstr);
      }
    }
  return TCL_OK;
}


/* LcompList returns list of compounds having the same right character as
   the character given as argument */
int
LcompList(ClientData clientData, Tcl_Interp *interp,
	  int argc, char **argv)
{
  int		i,
		givench,
		flevel;
  char		resultstr[120];

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    givench = atoi(argv[1]);
    flevel = atoi(argv[2]);
    simpFlag = strcmp(argv[3],"fanti");
    }

  /* we just return everything on the characters Lpart (left partner) list */
  Tcl_ResetResult(interp);
  for (i=0;i<chartbl[givench].Lpart.n_entries;i++) {
    if (comptbl[chartbl[givench].Lpart.ent[i]].freq < flevel) {
      construct_comp_entry(resultstr,
			   &(comptbl[chartbl[givench].Lpart.ent[i]]));
      Tcl_AppendElement(interp,resultstr);
      }
    }

  return TCL_OK;
}


/* Rcomplist returns list of compounds with same left character as given. */
int
RcompList(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
  int		i,
		givench,
		flevel;
  char		resultstr[120];

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    givench = atoi(argv[1]);
    flevel = atoi(argv[2]);
    simpFlag = strcmp(argv[3],"fanti");
    }

  /* we just return everything on the characters Rpart (right partner) list */
  Tcl_ResetResult(interp);
  for (i=0;i<chartbl[givench].Rpart.n_entries;i++) {
    if (comptbl[chartbl[givench].Rpart.ent[i]].freq < flevel) {
      construct_comp_entry(resultstr,
			   &(comptbl[chartbl[givench].Rpart.ent[i]]));
      Tcl_AppendElement(interp,resultstr);
      }
    }
  return TCL_OK;
}


/* radList() takes no arguments and returns a list of lists of, for each
   number of strokes, a space-separated pair for each radical of a character
   index for the first character containing the radical and the font string
   for the radical; (used for generating a radical->character look-up table) */
int
radList(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
#define MRS 11
  int	i,j, found;
  char	pair[22],
    	*tpos;
  char	*strokeList[MRS][50],	/* assume won't have >50 radicals per stroke */
	*ptrs[51];
  int	n_entries[MRS], ns;
  static char strokeTitles[MRS][5] = {
    "1","2","3","4","5","6","7","8","9","10",">10" };
  char	*tmpRes;	/* for Tcl dynamic allocation */
  
  if (clientData); /* shut compiler up */
  if (argc != 2)
    return TCL_ERROR;
  else
    simpFlag = strcmp(argv[1],"fanti");
  for (i=0;i<MRS;i++) n_entries[i] = 0;

  for (i=0;i<NRAD;i++) {
    /* find a character entry if there is one, or leave as 0 */
    for (found=0,j=3;j>-1;j--)
      if (radtbl[i].chars[j].n_entries > 0)
	found = radtbl[i].chars[j].ent[0];
    strcpy(pair,itoa(found));
    strcat(pair," ");
    tpos = pair + strlen(pair);

    /* copy in font string -- always simplified */
    strcpy(tpos,chartbl[radtbl[i].charidx].utf_jh);
    tpos += 3;
    *tpos = '\0';

    /* append entry to proper strokelist */
    ns = radtbl[i].strokes > 10 ? 10 : radtbl[i].strokes - 1;
    if (!(strokeList[ns][n_entries[ns]] = (char *) malloc(22 * sizeof(char))))
      memfail("stroke list");
    strcpy(strokeList[ns][n_entries[ns]],pair);
    n_entries[ns]++;
    }
  /* now build up list structure to return */
  Tcl_ResetResult(interp);  /* clear from previous */
  for (ns=0;ns<MRS;ns++) {
    ptrs[0] = strokeTitles[ns];
    for (i=0;i<n_entries[ns];i++)
      ptrs[i+1] = strokeList[ns][i];
    tmpRes = Tcl_Merge(n_entries[ns]+1,ptrs);
    Tcl_AppendElement(interp,tmpRes);
    Tcl_Free(tmpRes);
    for (i=0;i<n_entries[ns];i++)
      free(strokeList[ns][i]);
    }

  return(TCL_OK);
}


/* pyChar() returns the first character having the given pronunciation */
int
pyChar(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
  pinyinst	*pyt;
  int		flevel, pyidx, i, nEnt, res;

  if (clientData); /* shut compiler up */

  if (argc != 3)
    return TCL_ERROR;
  else {	/* parse pinyin string */
    pyidx = pin_index(argv[1]);
    flevel = atoi(argv[2]);
    if (pyidx < 0)
      return TCL_ERROR;
    else
      pyt = &(pinyintbl[pyidx]);
  }
  
  /* clear last and return new result */
  Tcl_ResetResult(interp);
  nEnt = pyt->chars.n_entries;
  if (nEnt > 0) {
    /* find first character satisfying our frequency criterion or use first */
    res = pyt->chars.ent[0];
    for (i=0; i < nEnt; i++) {
      if (chartbl[pyt->chars.ent[i]].freq < flevel) {
        res = pyt->chars.ent[i];
        break;
      }
    }
    Tcl_AppendElement(interp, itoa(res));
    return TCL_OK;
  }
  return TCL_ERROR;  /* didn't find anything */
}


/* randChar() returns a random character satisfying the frequency criterion */
int
randChar(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
  int		z, flevel;
  static char	chrIdx[10];

  if (clientData); /* shut compiler up */

  if (argc != 2)
    return TCL_ERROR;
  else
    flevel = atoi(argv[1]);

  do {
    z = (int) floor((float) nchars * dr());
    }
  while (chartbl[z].freq >= flevel);
  strcpy(chrIdx, itoa(z));
  Tcl_ResetResult(interp);
  Tcl_SetResult(interp, chrIdx, TCL_STATIC);

  return TCL_OK;
}


/* convertSelection() attempts to find a Chinese character as the first part
   of the passed-in string and convert the font index to a character index;
   if this fails, the passed-in index is returned as default */
int
convertSelection(ClientData clientData, Tcl_Interp *interp,
                 int argc, char **argv)
{
  uchar		chrstr[2], dummy1, dummy2;
  char		tmpUTFstr[4], tmpGBstr[4];
  int		default_idx, attempt_idx, gbIdx;
  Tcl_Encoding	b5, gb;
  int		srcRead,dstWroteB,dstWroteC;

  if (clientData); /* shut compiler up */

  if (argc != 4)
    return TCL_ERROR;
  else {
    /* get the first Chinese character out of the selection; Tcl inserts escape
       characters before each of the constituent bytes, which we drop */
    simpFlag = strcmp(argv[2],"fanti");
//fprintf(stderr,"simpFlag = %d\n",simpFlag);
    default_idx = atoi(argv[3]);
  }

//fprintf(stderr,"selection is \"%s\" of length %d\n",argv[1],strlen(argv[1]));

  /* case 1: string too short to have a GB character */
  if (strlen(argv[1]) < 2) {
    attempt_idx = default_idx;
//fprintf(stderr,"selection too short\n",simpFlag);
  } else
    sscanf(argv[1],"%c%c%c%c",&dummy1,&(chrstr[0]),&dummy2,&(chrstr[1]));

  /* PENDING: only works with GB */
  /* get the character's GB font index */
  if (simpFlag) {
    gbIdx = f_char2int(chrstr[0], chrstr[1]);
//fprintf(stderr,"gbIdx = %d\n",gbIdx);
    attempt_idx = fnt2chr[gbIdx];  /* remove after below fixed */
  } else {
    fprintf(stderr,"hanzim: selection conversion not supported yet for traditional (sorry!)\n");
    attempt_idx = -1;
    /* messy: convert to UTF-8 then to GB */
    /* ... b5 -> UTF */
    /* b5 = Tcl_GetEncoding(interp,"big5"); */   /* can't fail at this stage */
    /* Tcl_ExternalToUtf(interp,b5,chrstr,2,0,NULL,tmpUTFstr,4,
                      &srcRead,&dstWroteB,&dstWroteC);
    Tcl_FreeEncoding(b5);
    printf("convert: got UTFstring \"%s\" of length %d.\n",tmpUTFstr,
           strlen(tmpUTFstr)); */
    /* ... UTF -> GB */
    /* the following fails, and we need our own u2g function in here ... */
    /* gb = Tcl_GetEncoding(interp,"gb2312");
    if ((gb == NULL)) i_error("tcl gb2312 encoding not installed");
    Tcl_UtfToExternal(interp,gb,tmpUTFstr,strlen(tmpUTFstr),0,NULL,tmpGBstr,4,
                      &srcRead,&dstWroteB,&dstWroteC);
    Tcl_FreeEncoding(gb);
    printf("convert: got GBstring \"%s\" of length %d.\n",tmpGBstr,
           strlen(tmpGBstr));
    gbIdx = f_char2int((uchar)tmpGBstr[0], (uchar)tmpGBstr[1]);
    printf("convert: gbIdx = %d\n", gbIdx);
    printf("convert: attemptIdx = %d\n", fnt2chr[gbIdx]); */
    /* attempt_idx = fnt2chr[gbIdx]; */
  }

  /* attempt_idx = fnt2chr[gbIdx]; */  /* use after above fixed */
  if (attempt_idx == -1) {
    attempt_idx = default_idx;
//fprintf(stderr,"defaulting...\n");
  }

  /* clear last and return new result */
  Tcl_ResetResult(interp);
//fprintf(stderr,"return index = %d\n",attempt_idx);
  Tcl_AppendElement(interp, itoa(attempt_idx));
  return TCL_OK;

}


/* charinfo() returns a list of a given character's character string, pinyin,
   and definition */
int
charinfo(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  int		cr;
  characterst	*cht;
  char		chr[4],
		pinyin[20],
		defn[80];

  if (clientData); /* shut compiler up */

  if ((argc != 3) || ((cr = atoi(argv[1])) >= nchars) ||
           (chartbl[cr].fontidx == -1)) {
    return TCL_ERROR;
  } else {
    cht = &(chartbl[cr]);
    simpFlag = strcmp(argv[2],"fanti");
  }
  if (simpFlag) strcpy(chr,cht->utf_jh);
  else strcpy(chr,cht->utf_ft);

  /* clear last and return new result */
  Tcl_ResetResult(interp);
  /* character, pinyin, meaning */
  Tcl_AppendElement(interp,chr);
  strcpy(pinyin,init_str[pinyintbl[cht->pinyins.ent[0]].init]);
  strcat(pinyin,fin_str[pinyintbl[cht->pinyins.ent[0]].finl]);
  strcat(pinyin,itoa(pinyintbl[cht->pinyins.ent[0]].tone + 1));
  Tcl_AppendElement(interp,pinyin);
  defn[0] = '\0';
  catdef(defn,cht->meaning,18,2,0, 0);
  Tcl_AppendElement(interp,defn);
  /* radical, radical meaning */
  if ((cr = cht->radidx) >= 0) {
    strcpy(chr,chartbl[radtbl[cr].charidx].utf_jh);
    strncpy(defn,radtbl[cr].meaning,79);
    }
  else strcpy(chr,""), strcpy(defn,"");
  Tcl_AppendElement(interp,chr), Tcl_AppendElement(interp,defn);
  /* remainder, remainder pronunciation */
  if ((cr = cht->remainder) >= 0) {
    if (simpFlag) strcpy(chr,chartbl[cr].utf_jh);
    else strcpy(chr,chartbl[cr].utf_ft);
    strcpy(pinyin,init_str_t[pinyintbl[chartbl[cr].pinyins.ent[0]].init]);
    strcat(pinyin,fin_str_t[pinyintbl[chartbl[cr].pinyins.ent[0]].finl][pinyintbl[chartbl[cr].pinyins.ent[0]].tone]);
    }
  else strcpy(chr,""), strcpy(pinyin,"");
  Tcl_AppendElement(interp,chr), Tcl_AppendElement(interp,pinyin);
  /* main pronunciation but encoded in utf with diacriticals */
  strcpy(pinyin,init_str_t[pinyintbl[cht->pinyins.ent[0]].init]);
  strcat(pinyin,fin_str_t[pinyintbl[cht->pinyins.ent[0]].finl][pinyintbl[cht->pinyins.ent[0]].tone]);
  Tcl_AppendElement(interp,pinyin);
  return TCL_OK;
}


/* construct_char_entry() builds a tcl list of:
   character index, font code (in chars), pinyin string in gb font
   representation, pinyin string in alphanumeric, and meaning
   for a character given as argument, and returns it as a string in s */
void
construct_char_entry(char *s, int cidx)
{
  characterst	*cht;
  char		tidx[10], tchar[10], tpyaln[20], tpyfnt[20], tmeaning[DEFNL],
		*ptrs[5];
  char		*tmpRes;	/* for Tcl dynamic allocation */

  ptrs[0] = tidx, ptrs[1] = tchar, ptrs[2] = tpyfnt, ptrs[3] = tpyaln,
    ptrs[4] = tmeaning;

  cht = &(chartbl[cidx]);
  
  strcpy(tidx,itoa(cidx));
  if (simpFlag) strcpy(tchar,cht->utf_jh);
  else strcpy(tchar,cht->utf_ft);
  strcpy(tpyaln,init_str[pinyintbl[cht->pinyins.ent[0]].init]);
  strcat(tpyaln,fin_str[pinyintbl[cht->pinyins.ent[0]].finl]);
  strcat(tpyaln,itoa(pinyintbl[cht->pinyins.ent[0]].tone + 1));
  strcpy(tmeaning,"\t");
  catdef(tmeaning,cht->meaning,20,2,1, 1);
  strcpy(tpyfnt,init_str_t[pinyintbl[cht->pinyins.ent[0]].init]);
  strcat(tpyfnt,fin_str_t[pinyintbl[cht->pinyins.ent[0]].finl][pinyintbl[cht->pinyins.ent[0]].tone]);
  tmpRes = Tcl_Merge(5,ptrs);
  strcpy(s, tmpRes);
  Tcl_Free(tmpRes);

}


/* construct_comp_entry() composes a tcl list of:
   character indices, font strings, hyphenated pinyin strings, and
   compound meaning for a compound given as argument, and returns it as
   a string in s */
void
construct_comp_entry(char *s, compoundst *comp)
{
  characterst	*cht1, *cht2;
  char		tidx[20], tchar[20], pystrL[20], pystrR[20], tmeaning[DEFNL],
		*ptrs[5];
  char		*tmpRes;	/* for Tcl dynamic allocation */

  ptrs[0]=tidx, ptrs[1]=tchar, ptrs[2]=pystrL, ptrs[3]=pystrR,
    ptrs[4]=tmeaning;

  cht1 = &(chartbl[comp->Lcharidx]), cht2 = &(chartbl[comp->Rcharidx]);

  strcpy(tidx,itoa(comp->Lcharidx)), strcat(tidx," ");
  strcat(tidx,itoa(comp->Rcharidx)), strcat(tidx," ");
  if (simpFlag)
    strcpy(tchar,cht1->utf_jh), strcat(tchar,cht2->utf_jh);
  else
    strcpy(tchar,cht1->utf_ft), strcat(tchar,cht2->utf_ft);
  /* pronunciations encoded in utf with diacriticals;
     each pinyin character consumes 2 chars, for a max of 12 */
  strcpy(pystrL,init_str_t[pinyintbl[cht1->pinyins.ent[0]].init]);
  strcat(pystrL,fin_str_t[pinyintbl[cht1->pinyins.ent[0]].finl][pinyintbl[cht1->pinyins.ent[0]].tone]);
  strcpy(pystrR,init_str_t[pinyintbl[cht2->pinyins.ent[0]].init]);
  strcat(pystrR,fin_str_t[pinyintbl[cht2->pinyins.ent[0]].finl][pinyintbl[cht2->pinyins.ent[0]].tone]);
  strcpy(tmeaning,"");
  catdef(tmeaning,comp->meaning,31,2,0, 1);
  tmpRes = Tcl_Merge(5,ptrs);
  strcpy(s,tmpRes);
  Tcl_Free(tmpRes);

}


/* catdef() copies nl lines of length llen from definition string src to a
   destination string dest, wrapping words and padding all lines with
   spaces if necessary */
void
catdef(char *dest, char *src, int llen, int nl, int ntabs, char uscoreflag)
{
  int 	l, c;				/* indices */
  char	*spos, *dpos, *tspos, *last;	/* character pointers */
  
  last = src + strlen(src) - 1;
  spos = src, dpos = dest + strlen(dest);
  for (l=0; l<nl;l++) {	/* for each line */
    /* case 1: defn ended already - fully pad with spaces */
    if (spos > last) {
      dpos += spacepad(dpos,llen);
      }
    /* case 2: defn will end this line - copy and pad with spaces */
    else if ((spos + llen) > last) {
      for (c=0; spos <= last; c++)
	*dpos++ = *spos++;
      dpos += spacepad(dpos, llen - c);
      spos = last + 1;
      }
    /* case 3: line ends on a word in definition - copy up to word and pad */
    /* PENDING: fails where single word > line length */
    else if (!isspace(*(spos+llen-1)) && !isspace(*(spos+llen))) {
      for (tspos = spos + llen; !isspace(*tspos); tspos--);
      for (c=0; spos <= tspos; c++)
	*dpos++ = *spos++;
      dpos += spacepad(dpos, llen - c); /* - 1); */
      }
    /* case 4: line ends on a space or end of a word - copy and don't pad */
    else {
      for (c=0;c<llen;c++)
	*dpos++ = *spos++;
      if (*spos == ' ') spos++;	/* swallow space at beginning of next line */
      }

    if (l < (nl-1)) {
      *dpos++ = '\n';
      if (uscoreflag) *dpos++ = '_';
      for (c=0;c<ntabs;c++) *dpos++ = '\t';
      *dpos++ = ' '; /* indent lines after first */
      if (l == 0) llen--;
      }

    }

  /* add line marker and null terminate */
  if (uscoreflag && isspace(*(dpos-1))) *(dpos-1) = '_';
  *dpos = '\0';

}

/* issep() returns nonzero if space, / or - line-ending OK */
int
issep(char c)
{
  return isspace(c) || (c == '/') || (c == '-');
}

/* spacepad() adds n spaces to string dest */
int
spacepad(char *dest, int n)
{
  int	i;
  char	*tpos;

  tpos = dest;
  for (i=0;i<n;i++)
    *tpos++ = ' ';

  return n;

}

/* yingHan returns a chinese string for use as a button title -- don't know
   how to represent it straight in tcl */
int
yingHanCh(ClientData clientData, Tcl_Interp *interp,int argc, char **argv)
{
  static char s[20];

  if (clientData); /* shut compiler up */
  if (argc); /* shut compiler up */
  if (argv); /* shut compiler up */

  sprintf(s,"%s%s",chartbl[fnt2chr[54178]].utf_jh,
	  chartbl[fnt2chr[47802]].utf_jh);
  Tcl_ResetResult(interp);
  Tcl_SetResult(interp,s, TCL_STATIC);
  return TCL_OK;
}

/* searchDefn searches the compounds database for any with definitions
   containing the *entire string* given and returns them in the same fashion
   as [LR]compList */
int
searchDefn(ClientData clientData, Tcl_Interp *interp,
	  int argc, char **argv)
{
  int		i;
  char		resultstr[120],
		*defn;

  if (clientData); /* shut compiler up */

  if (argc != 2) return TCL_ERROR;
  else {
    defn = argv[1];
    }

  /* clear last and return new result */
  Tcl_ResetResult(interp);
  /* need to search through entire compound database.. note, using strstr()
     might lead to compatibility problems :( */
  /* PENDING: try to match only whole words? */
  for (i=0;i<Ncomp;i++) {
    if (strstr(comptbl[i].meaning,defn)) {
      construct_comp_entry(resultstr,&(comptbl[i]));
      Tcl_AppendElement(interp,resultstr);
      }
    }
  
  return TCL_OK;
}
